package xyz.zhouxy.plusone.validator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.regex.Pattern;

import com.google.common.base.Strings;

import xyz.zhouxy.plusone.commons.util.Predicates;
import xyz.zhouxy.plusone.commons.util.RegexUtil;

public class BaseValidator<T> {
    private final List<Consumer<T>> rules = new ArrayList<>();
    private final List<PropertyValidator<T, ?, ?>> propertyValidators = new ArrayList<>();

    protected void withRule(final Predicate<T> rule, final String errorMessage) {
        withRule(rule, () -> new InvalidInputException(errorMessage));
    }

    protected <E extends RuntimeException> void withRule(Predicate<T> rule, Supplier<E> exceptionBuilder) {
        withRule(rule, value -> exceptionBuilder.get());
    }

    protected <E extends RuntimeException> void withRule(Predicate<T> condition,
            Function<T, E> exceptionBuilder) {
        withRule(value -> {
            if (!condition.test(value)) {
                throw exceptionBuilder.apply(value);
            }
        });
    }

    protected void withRule(Consumer<T> rule) {
        this.rules.add(rule);
    }

    protected final <R> ObjectValidator<T, R> ruleFor(Function<T, R> getter) {
        ObjectValidator<T, R> validValueHolder = new ObjectValidator<>(getter);
        propertyValidators.add(validValueHolder);
        return validValueHolder;
    }

    protected final IntValidator<T> ruleForInt(Function<T, ?> getter) {
        IntValidator<T> validValueHolder = new IntValidator<>(getter);
        propertyValidators.add(validValueHolder);
        return validValueHolder;
    }

    protected final DoubleValidator<T> ruleForDouble(Function<T, ?> getter) {
        DoubleValidator<T> validValueHolder = new DoubleValidator<>(getter);
        propertyValidators.add(validValueHolder);
        return validValueHolder;
    }

    protected final BoolValidator<T> ruleForBool(Function<T, ?> getter) {
        BoolValidator<T> validValueHolder = new BoolValidator<>(getter);
        propertyValidators.add(validValueHolder);
        return validValueHolder;
    }

    protected final StringValidator<T> ruleForString(Function<T, ?> getter) {
        StringValidator<T> validValueHolder = new StringValidator<>(getter);
        propertyValidators.add(validValueHolder);
        return validValueHolder;
    }

    protected final <E> CollectionValidator<T, E> ruleForCollection(Function<T, Collection<E>> getter) {
        CollectionValidator<T, E> validValueHolder = new CollectionValidator<>(getter);
        propertyValidators.add(validValueHolder);
        return validValueHolder;
    }

    public void validate(T obj) {
        for (Consumer<T> rule : this.rules) {
            rule.accept(obj);
        }
        for (PropertyValidator<T, ?, ?> valueValidator : this.propertyValidators) {
            valueValidator.validate(obj);
        }
    }

    abstract static class PropertyValidator<DTO, PROPERTY, THIS> {
        Function<DTO, ?> getter;
        Validator<PROPERTY> validator = new Validator<>();

        PropertyValidator(Function<DTO, ?> getter) {
            this.getter = getter;
        }

        <E extends RuntimeException> void withRule(Predicate<? super PROPERTY> condition,
                Function<PROPERTY, E> exceptionCreator) {
            withRule(value -> {
                if (!condition.test(value)) {
                    throw exceptionCreator.apply(value);
                }
            });
        }

        private void withRule(Consumer<PROPERTY> rule) {
            this.validator.addRule(rule);
        }

        // ====================
        // ====== Object ======
        // ====================

        // ====== notNull =====

        public THIS notNull() {
            return notNull("Value could not be null.");
        }

        public THIS notNull(String errMsg) {
            return notNull(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> THIS notNull(Supplier<E> exceptionCreator) {
            return notNull(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> THIS notNull(Function<PROPERTY, E> exceptionCreator) {
            withRule(Objects::nonNull, exceptionCreator);
            return thisObject();
        }

        // ====== isNull =====

        public THIS isNull(String errMsg) {
            return isNull(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> THIS isNull(Supplier<E> exceptionCreator) {
            return isNull(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> THIS isNull(Function<PROPERTY, E> exceptionCreator) {
            withRule(Objects::isNull, exceptionCreator);
            return thisObject();
        }

        // ===== equals =====

        public THIS equalsThat(Object that) {
            return equalsThat(that, value -> new InvalidInputException("(%s) 必须与 (%s) 相等", value, that));
        }

        public THIS equalsThat(Object that, String errMsg) {
            return equalsThat(that, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> THIS equalsThat(
                Object that, Supplier<E> exceptionCreator) {
            return equalsThat(that, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> THIS equalsThat(
                Object that, Function<PROPERTY, E> exceptionCreator) {
            withRule(value -> Objects.equals(value, that), exceptionCreator);
            return thisObject();
        }

        // ===== isTrue =====

        public THIS isTrue(Predicate<? super PROPERTY> condition) {
            return isTrue(condition, "无效的用户输入");
        }

        public THIS isTrue(Predicate<? super PROPERTY> condition, String errMsg) {
            return isTrue(condition, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> THIS isTrue(
                Predicate<? super PROPERTY> condition,
                Supplier<E> exceptionCreator) {
            return isTrue(condition, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> THIS isTrue(
                Predicate<? super PROPERTY> condition,
                Function<PROPERTY, E> exceptionCreator) {
            withRule(condition, exceptionCreator);
            return thisObject();
        }

        // ===== isTrue =====

        public THIS isTrue(Collection<Predicate<PROPERTY>> conditions) {
            return isTrue(conditions, "无效的用户输入");
        }

        public THIS isTrue(Collection<Predicate<PROPERTY>> conditions, String errMsg) {
            return isTrue(conditions, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> THIS isTrue(
                Collection<Predicate<PROPERTY>> conditions,
                Supplier<E> exceptionCreator) {
            return isTrue(conditions, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> THIS isTrue(
                Collection<Predicate<PROPERTY>> conditions,
                Function<PROPERTY, E> exceptionCreator) {
            for (Predicate<PROPERTY> condition : conditions) {
                withRule(condition, exceptionCreator);
            }
            return thisObject();
        }

        // ========================================================================

        @SuppressWarnings("unchecked")
        void validate(DTO obj) {
            PROPERTY value = (PROPERTY) this.getter.apply(obj);
            this.validator.validate(value);
        }

        static <V> Function<V, InvalidInputException> convertExceptionCreator(String errMsg) {
            return value -> new InvalidInputException(errMsg);
        }

        static <V, E extends RuntimeException> Function<V, E> convertExceptionCreator(
                Supplier<E> exceptionSupplier) {
            return value -> exceptionSupplier.get();
        }

        protected abstract THIS thisObject();
    }

    public static class BoolValidator<DTO> extends PropertyValidator<DTO, Boolean, BoolValidator<DTO>> {

        BoolValidator(Function<DTO, ?> getter) {
            super(getter);
        }

        // ====== isTrue ======

        public BoolValidator<DTO> isTrue() {
            return isTrue("The value must be true.");
        }

        public BoolValidator<DTO> isTrue(String errMsg) {
            return isTrue(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> BoolValidator<DTO> isTrue(Supplier<E> exceptionCreator) {
            return isTrue(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> BoolValidator<DTO> isTrue(
                Function<Boolean, E> exceptionCreator) {
            super.withRule(Boolean.TRUE::equals, exceptionCreator);
            return this;
        }

        // ====== isFalse ======

        public BoolValidator<DTO> isFalse() {
            return isFalse("The value must be false.");
        }

        public BoolValidator<DTO> isFalse(String errMsg) {
            return isFalse(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> BoolValidator<DTO> isFalse(Supplier<E> exceptionCreator) {
            return isFalse(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> BoolValidator<DTO> isFalse(
                Function<Boolean, E> exceptionCreator) {
            super.withRule(Boolean.FALSE::equals, exceptionCreator);
            return this;
        }

        @Override
        protected BoolValidator<DTO> thisObject() {
            return this;
        }
    }

    public static class DoubleValidator<DTO> extends PropertyValidator<DTO, Double, DoubleValidator<DTO>> {

        DoubleValidator(Function<DTO, ?> getter) {
            super(getter);
        }

        public DoubleValidator<DTO> between(double min, double max) {
            return between(min, max, String.format("数值不在 %s 和 %s 之间", String.valueOf(min), String.valueOf(max)));
        }

        public DoubleValidator<DTO> between(double min, double max, String errMsg) {
            return between(min, max, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> DoubleValidator<DTO> between(double min, double max,
                Supplier<E> exceptionCreator) {
            return between(min, max, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> DoubleValidator<DTO> between(double min, double max,
                Function<Double, E> exceptionCreator) {
            super.withRule(value -> (value >= min && value < max), exceptionCreator);
            return this;
        }

        @Override
        protected DoubleValidator<DTO> thisObject() {
            return this;
        }
    }

    public static class IntValidator<DTO> extends PropertyValidator<DTO, Integer, IntValidator<DTO>> {

        IntValidator(Function<DTO, ?> getter) {
            super(getter);
        }

        public IntValidator<DTO> between(int min, int max) {
            return between(min, max, String.format("数值不在 %d 和 %d 之间", min, max));
        }

        public IntValidator<DTO> between(int min, int max, String errMsg) {
            return between(min, max, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> IntValidator<DTO> between(int min, int max,
                Supplier<E> exceptionCreator) {
            return between(min, max, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> IntValidator<DTO> between(int min, int max,
                Function<Integer, E> exceptionCreator) {
            super.withRule(value -> (value >= min && value < max), exceptionCreator);
            return this;
        }

        @Override
        protected IntValidator<DTO> thisObject() {
            return this;
        }
    }

    public static class StringValidator<DTO> extends PropertyValidator<DTO, String, StringValidator<DTO>> {

        StringValidator(Function<DTO, ?> getter) {
            super(getter);
        }

        // ====================
        // ====== String ======
        // ====================

        // ===== matches =====

        public StringValidator<DTO> matches(Pattern regex, String errMsg) {
            return matches(regex, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> matches(
                Pattern regex,
                Supplier<E> exceptionCreator) {
            return matches(regex, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> matches(
                Pattern regex,
                Function<String, E> exceptionCreator) {
            super.withRule(input -> RegexUtil.matches(input, regex), exceptionCreator);
            return this;
        }

        // ===== matchesOne =====

        public StringValidator<DTO> matchesOne(Pattern[] regexs, String errMsg) {
            return matchesOne(regexs, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesOne(
                Pattern[] regexs,
                Supplier<E> exceptionCreator) {
            return matchesOne(regexs, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesOne(
                Pattern[] regexs,
                Function<String, E> exceptionCreator) {
            super.withRule(input -> RegexUtil.matchesOne(input, regexs), exceptionCreator);
            return this;
        }

        public StringValidator<DTO> matchesOne(List<Pattern> regexs, String errMsg) {
            return matchesOne(regexs, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesOne(
                List<Pattern> regexs,
                Supplier<E> exceptionCreator) {
            return matchesOne(regexs, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesOne(
                List<Pattern> regexs,
                Function<String, E> exceptionCreator) {
            super.withRule(input -> RegexUtil.matchesOne(input, regexs.toArray(new Pattern[regexs.size()])),
                    exceptionCreator);
            return this;
        }

        // ===== matchesAll =====

        public StringValidator<DTO> matchesAll(Pattern[] regexs, String errMsg) {
            return matchesAll(regexs, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesAll(
                Pattern[] regexs,
                Supplier<E> exceptionCreator) {
            return matchesAll(regexs, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesAll(
                Pattern[] regexs,
                Function<String, E> exceptionCreator) {
            super.withRule(input -> RegexUtil.matchesAll(input, regexs), exceptionCreator);
            return this;
        }

        public StringValidator<DTO> matchesAll(Collection<Pattern> regexs, String errMsg) {
            return matchesAll(regexs, convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesAll(
                Collection<Pattern> regexs,
                Supplier<E> exceptionCreator) {
            return matchesAll(regexs, convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> matchesAll(
                Collection<Pattern> regexs,
                Function<String, E> exceptionCreator) {
            super.withRule(input -> RegexUtil.matchesAll(input, regexs.toArray(new Pattern[regexs.size()])),
                    exceptionCreator);
            return this;
        }

        // ====== notEmpty =====

        public StringValidator<DTO> notEmpty() {
            return notEmpty("The string argument must not be empty.");
        }

        public StringValidator<DTO> notEmpty(String errMsg) {
            return notEmpty(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> notEmpty(Supplier<E> exceptionCreator) {
            return notEmpty(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> notEmpty(
                Function<String, E> exceptionCreator) {
            super.withRule(Predicates.not(Strings::isNullOrEmpty), exceptionCreator);
            return this;
        }

        // ====== isEmpty =====

        public StringValidator<DTO> isEmpty() {
            return notEmpty("The string argument must be empty.");
        }

        public StringValidator<DTO> isEmpty(String errMsg) {
            return isEmpty(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> StringValidator<DTO> isEmpty(Supplier<E> exceptionCreator) {
            return isEmpty(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> StringValidator<DTO> isEmpty(
                Function<String, E> exceptionCreator) {
            super.withRule(Strings::isNullOrEmpty, exceptionCreator);
            return this;
        }

        @Override
        protected StringValidator<DTO> thisObject() {
            return this;
        }
    }

    public static class ObjectValidator<DTO, T> extends PropertyValidator<DTO, T, ObjectValidator<DTO, T>> {

        ObjectValidator(Function<DTO, T> getter) {
            super(getter);
        }

        @Override
        protected ObjectValidator<DTO, T> thisObject() {
            return this;
        }
    }

    public static class CollectionValidator<DTO, T>
            extends PropertyValidator<DTO, Collection<T>, CollectionValidator<DTO, T>> {

        CollectionValidator(Function<DTO, Collection<T>> getter) {
            super(getter);
        }

        // ====== notEmpty =====

        public CollectionValidator<DTO, T> notEmpty(String errMsg) {
            return notEmpty(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> CollectionValidator<DTO, T> notEmpty(Supplier<E> exceptionCreator) {
            return notEmpty(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> CollectionValidator<DTO, T> notEmpty(
                Function<Collection<T>, E> exceptionCreator) {
            super.withRule(value -> value != null && !value.isEmpty(), exceptionCreator);
            return this;
        }

        // ====== isEmpty =====

        public CollectionValidator<DTO, T> isEmpty(String errMsg) {
            return isEmpty(convertExceptionCreator(errMsg));
        }

        public <E extends RuntimeException> CollectionValidator<DTO, T> isEmpty(Supplier<E> exceptionCreator) {
            return isEmpty(convertExceptionCreator(exceptionCreator));
        }

        public <E extends RuntimeException> CollectionValidator<DTO, T> isEmpty(
                Function<Collection<T>, E> exceptionCreator) {
            super.withRule(value -> value == null || value.isEmpty(), exceptionCreator);
            return this;
        }

        @Override
        protected CollectionValidator<DTO, T> thisObject() {
            return this;
        }
    }

}
